# Most of the collecting code was picked from other places
from __future__ import absolute_import

import logging
from typing import Any, Dict, Optional

from mlfoundry.background import utils as bg_utils
from mlfoundry.background.interface import Interface
from mlfoundry.vendor.pynvml import pynvml

logger = logging.getLogger(__name__)

try:
    import psutil
except ImportError:
    logger.warning(
        f"`psutil` is not installed. System metrics will not be reported. Install with `pip install psutil`"
    )
    psutil = None

# TODO (chiragjn): Add cgroup reading mode for Linux/Docker and TPU support
MEGABYTE = float(1024 * 1024)


def gpu_in_use_by_this_process(gpu_handle) -> bool:
    if not psutil:
        return False
    base_process = psutil.Process().parent() or psutil.Process()
    our_processes = base_process.children(recursive=True)
    our_processes.append(base_process)
    our_pids = {process.pid for process in our_processes}
    compute_pids = {
        process.pid
        for process in pynvml.nvmlDeviceGetComputeRunningProcesses(gpu_handle)
    }
    graphics_pids = {
        process.pid
        for process in pynvml.nvmlDeviceGetGraphicsRunningProcesses(gpu_handle)
    }
    pids_using_device = compute_pids | graphics_pids
    return len(pids_using_device & our_pids) > 0


# TODO (chiragjn): Separate stats collecting functionality into another module for unit testing
class SystemMetricsJob(bg_utils.BackgroundJob):
    """
    Collect system resources (cpu, memory, disk, network, gpu) metrics and log them with the run

    Attributes:
        num_samples_to_aggregate: number of samples to aggregate before logging them
        gpu_count: number of GPUs detected
        _pid: process PID to monitor and collect stats for
        _interface: `background.interface.Interface` instance to publish collected stats with
        _sampled: dict holding metric name to list of samples for that metric
        _sample_count: how many samples out of `num_samples_to_aggregate`  have been collected
        _network_init: number of bytes sent and received before we start measuring for our run
    """

    name_prefix = "MLFSystemMetricsThread"

    def __init__(
        self,
        pid: int,
        interface: Interface,
        num_samples_to_aggregate: int = 15,
        interval: float = 2.0,
    ) -> None:
        """
        Args:
            pid: process PID to monitor and collect stats for
            interface: `background.interface.Interface` instance to publish collected stats with
            num_samples_to_aggregate: number of samples to aggregate before logging them, defaults to 15
            interval: time in seconds to sleep between consecutive stats collecting, defaults to 2 seconds
        """
        super().__init__(
            name=f"{SystemMetricsJob.name_prefix}:{interface.run_id}", interval=interval
        )
        self.num_samples_to_aggregate = num_samples_to_aggregate
        try:
            pynvml.nvmlInit()
            self.gpu_count = pynvml.nvmlDeviceGetCount()
        except pynvml.NVMLError:
            self.gpu_count = 0
        self._pid = pid
        self._interface: Optional[Interface] = interface
        self._sampled = {}
        self._sample_count = 0
        if psutil:
            net = psutil.net_io_counters()
            self._network_init = {"sent": net.bytes_sent, "recv": net.bytes_recv}
        else:
            self._network_init = {"sent": 0, "recv": 0}
            logger.warning(
                "`psutil` not installed, only GPU stats will be reported. Install with `pip install psutil`"
            )

    def _collect_gpu_stats(self, stats: Dict[str, Any]) -> None:
        """
        Collect GPU specific stats in given `stats` dict

        Args:
            stats: dict to collect stats into
        """
        for i in range(0, self.gpu_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            try:
                utilz = pynvml.nvmlDeviceGetUtilizationRates(handle)
                memory = pynvml.nvmlDeviceGetMemoryInfo(handle)
                temp = pynvml.nvmlDeviceGetTemperature(
                    handle, pynvml.NVML_TEMPERATURE_GPU
                )
                in_use_by_current_process = gpu_in_use_by_this_process(handle)

                stats[f"gpu.{i}.gpu"] = utilz.gpu
                stats[f"gpu.{i}.memory"] = utilz.memory
                stats[f"gpu.{i}.memory_allocated.percent"] = (
                    memory.used / float(memory.total)
                ) * 100
                stats[f"gpu.{i}.temperature"] = temp

                if in_use_by_current_process:
                    stats[f"gpu.process.{i}.gpu"] = utilz.gpu
                    stats[f"gpu.process.{i}.memory"] = utilz.memory
                    stats[f"gpu.process.{i}.memory_allocated.percent"] = (
                        memory.used / float(memory.total)
                    ) * 100
                    stats[f"gpu.process.{i}.temperature"] = temp

                try:
                    power_watts = pynvml.nvmlDeviceGetPowerUsage(handle) / 1000.0
                    power_capacity_watts = (
                        pynvml.nvmlDeviceGetEnforcedPowerLimit(handle) / 1000.0
                    )
                    power_usage = (power_watts / power_capacity_watts) * 100
                    stats[f"gpu.{i}.power.watts"] = power_watts
                    stats[f"gpu.{i}.power.percent"] = power_usage

                    if in_use_by_current_process:
                        stats[f"gpu.process.{i}.power.watts"] = power_watts
                        stats[f"gpu.process.{i}.power.percent"] = power_usage

                except pynvml.NVMLError:
                    pass
            except pynvml.NVMLError:
                pass

    def _collect_system_stats(self, stats: Dict[str, Any]):
        """
        Collect CPU, memory, disk and network related stats given `stats` dict

        Args:
            stats: dict to collect stats into
        """
        if psutil:
            net = psutil.net_io_counters()
            sysmem = psutil.virtual_memory()
            stats["cpu.percent"] = psutil.cpu_percent()
            stats["memory.percent"] = sysmem.percent
            stats["network.sent.bytes"] = net.bytes_sent - self._network_init["sent"]
            stats["network.recv.bytes"] = net.bytes_recv - self._network_init["recv"]
            stats["disk.percent"] = psutil.disk_usage("/").percent
            stats["proc.memory.available.mb"] = sysmem.available / MEGABYTE
            try:
                proc = psutil.Process(pid=self._pid)
                stats["proc.memory.rss.mb"] = proc.memory_info().rss / MEGABYTE
                stats["proc.memory.percent"] = proc.memory_percent()
                stats["proc.cpu.threads"] = proc.num_threads()
            except psutil.NoSuchProcess:
                pass

    def collect_stats(self) -> Dict[str, float]:
        """
        Collect gpu and system stats

        Returns:
            Dict[str, float]: dict mapping metric name to value for that metric
        """
        stats = {}
        self._collect_gpu_stats(stats)
        self._collect_system_stats(stats)
        return stats

    def _flush(self) -> None:
        """
        Aggregate (mean, max) and log collected stats so far and reset sampling state
        """
        if not self._sampled:
            return
        stats = self.collect_stats()
        metric_dict = {}
        for stat, value in stats.items():
            if isinstance(value, (float, int)):
                samples = list(self._sampled.get(stat, [value]))
                if samples:
                    metric_dict[f"mlf.system_metrics.{stat}.max"] = round(
                        max(samples), 2
                    )
                    metric_dict[f"mlf.system_metrics.{stat}.avg"] = round(
                        sum(samples) / len(samples), 2
                    )
        if self._interface:
            self._interface.log_metrics(metric_dict=metric_dict)
        self._sample_count = 0
        self._sampled = {}

    def _loop(self) -> None:
        """
        Collect stats and add them to samples. Flush if we have collected the max number of samples allowed
        """
        stats = self.collect_stats()
        for stat, value in stats.items():
            if isinstance(value, (int, float)):
                self._sampled[stat] = self._sampled.get(stat, [])
                self._sampled[stat].append(value)
        self._sample_count += 1
        if self._sample_count >= self.num_samples_to_aggregate:
            self._flush()

    def _finish(self) -> None:
        """
        Flush any remaining collected stats samples
        """
        self._flush()
        self._interface = None



from mlfoundry.log_types.dataset_artifact import (
    DatasetArtifact,
    DatasetArtifactActuals,
    DatasetArtifactActualsProfile,
    DatasetArtifactFeatures,
    DatasetArtifactFeaturesProfile,
    DatasetArtifactPredictions,
    DatasetArtifactPredictionsProfile,
)
from mlfoundry.log_types.dataset_schema import DatasetSchemaRunLog
from mlfoundry.log_types.dataset_stats import DatasetStatsRunLog, StatsSource
from mlfoundry.log_types.model_artifact import ModelArtifactRunLog



from mlflow.utils.rest_utils import MlflowHostCreds, http_request_safe

from mlfoundry.tracking.entities import TenantInfo


class AuthService:
    def __init__(self, tenant_info: TenantInfo):
        self.host_creds = MlflowHostCreds(host=tenant_info.auth_server_url)
        self.tenant_id = tenant_info.tenant_id

    def get_token(self, api_key: str) -> str:
        response = http_request_safe(
            host_creds=self.host_creds,
            endpoint="/api/v1/oauth/api-keys/token",
            method="post",
            json={"apiKey": api_key, "clientId": self.tenant_id},
        )
        response = response.json()
        token = response["accessToken"]
        return token



import atexit
import datetime
import json
import logging
import os
import queue
import time
import weakref
from typing import Any, Collection, Dict, Iterable, List, Optional, Union

import mlflow
import numpy as np
import pandas as pd
import whylogs
from mlflow.entities import Metric, Param, RunStatus, RunTag
from mlflow.tracking import MlflowClient

from mlfoundry import amplitude, constants, enums
from mlfoundry.background.interface import Interface
from mlfoundry.background.sender import SenderJob
from mlfoundry.background.system_metrics import SystemMetricsJob
from mlfoundry.dataset import DataSet, TabularDatasetDriver
from mlfoundry.exceptions import MlflowException, MlFoundryException
from mlfoundry.git_info import GitInfo
from mlfoundry.internal_namespace import NAMESPACE
from mlfoundry.metrics.v1 import get_metrics_calculator as get_metrics_calculator_v1
from mlfoundry.metrics.v2 import ComputedMetrics
from mlfoundry.metrics.v2 import get_metrics_calculator as get_metrics_calculator_v2
from mlfoundry.model import ModelDriver
from mlfoundry.run_utils import (
    NumpyEncoder,
    ParamsType,
    log_artifact_blob,
    process_params,
)
from mlfoundry.schema import Schema

logger = logging.getLogger(__name__)


class MlFoundryRun:
    # TODO (nikunjbjj): Seems like these artifact locations are defined in constants.py for loading.
    #  Need to change this.
    S3_DATASET_PATH = "datasets"
    S3_STATS_PATH = "stats"
    S3_WHYLOGS_PATH = "whylogs"
    S3_METRICS_PATH = "multi_dimensional_metrics"

    def __init__(
        self, experiment_id: str, run_id: str, log_system_metrics: bool = False
    ):
        self.experiment_id = str(experiment_id)
        mlflow.get_experiment(self.experiment_id)
        self.run_id = run_id
        # TODO (chiragjn): Shouldn't mlflow_client be a protected/private member?
        self.mlflow_client = MlflowClient()
        self._dataset_module: TabularDatasetDriver = TabularDatasetDriver(
            mlflow_client=self.mlflow_client, run_id=run_id
        )
        self._model_driver: ModelDriver = ModelDriver(
            mlflow_client=self.mlflow_client, run_id=run_id
        )
        # TODO (chiragjn): Make a settings module and allow enabling/disabling collection and changing intervals
        if log_system_metrics:
            # Interface and Sender do not belong under this condition but for now we don't need to init them otherwise
            self._interface = Interface(
                run_id=self.run_id,
                event_queue=queue.Queue(),
            )
            self._sender_job = SenderJob(
                interface=self._interface,
                mlflow_client=self.mlflow_client,
                interval=0.0,
            )
            self._system_metrics_job = SystemMetricsJob(
                pid=os.getpid(),
                interface=self._interface,
                num_samples_to_aggregate=15,
                interval=2.0,
            )
        else:
            self._interface = None
            self._sender_job = None
            self._system_metrics_job = None
        self._terminate_called = False
        self._start()

    def __repr__(self) -> str:
        return f"<{type(self).__name__} at 0x{id(self):x}: run_id={self.run_id!r}>"

    def __enter__(self):
        return self

    def _stop_background_jobs(self):
        """
        Stop launched background jobs (system metrics, sender) if any and try to finish them gracefully
        """
        # expect that this function can be called more than once in non-ideal scenarios and defend for it
        if not any([self._interface, self._system_metrics_job, self._sender_job]):
            return

        logger.info(
            f"Shutting down background jobs and syncing data for run with id {self.run_id!r}, "
            f"please don't kill this process..."
        )
        # Stop event producers
        if self._system_metrics_job:
            self._system_metrics_job.stop(timeout=2)
        # Stop accepting more events
        if self._interface:
            self._interface.close()
        # Finish consuming whatever is left
        if self._sender_job:
            try:
                self._sender_job.stop(disable_sleep=True, timeout=10)
            except KeyboardInterrupt:
                # TODO (chiragjn): Separate internal logging and stream to show messages to user
                print(
                    "Ctrl-C interrupt detected, background jobs are still terminating. "
                    "Press Ctrl-C again to stop."
                )
                self._sender_job.stop(disable_sleep=True, timeout=10)
        self._system_metrics_job = None
        self._sender_job = None
        self._interface = None

    def _terminate_run_if_running(self, termination_status: RunStatus):
        if self._terminate_called:
            return

        # Prevent double execution for termination
        self._terminate_called = True

        current_status = None
        try:
            # TODO (chiragjn): This is a network call, should we just cache it on the instance?
            current_status = self.mlflow_client.get_run(self.run_id).info.status
        except Exception as e:
            logger.warning(f"failed to get run status {termination_status} due to {e}")

        try:
            self._stop_background_jobs()
            # we do not need to set any termination status unless the run was in RUNNING state
            if current_status != RunStatus.to_string(RunStatus.RUNNING):
                return
            self.mlflow_client.set_terminated(
                self.run_id, RunStatus.to_string(termination_status)
            )
        except Exception as e:
            logger.warning(
                f"failed to set termination status {termination_status} due to {e}"
            )

    def __exit__(self, exc_type, exc_val, exc_tb):
        status = RunStatus.FINISHED if exc_type is None else RunStatus.FAILED
        self._terminate_run_if_running(status)

    def __del__(self):
        # TODO (chiragjn): Should this be marked as FINISHED or KILLED?
        self._terminate_run_if_running(RunStatus.FINISHED)

    def _start(self):
        def terminate_run_if_running_with_weakref(
            mlf_run_weakref: "weakref.ReferenceType[MlFoundryRun]",
            termination_status: RunStatus,
        ):
            _run = mlf_run_weakref()
            if _run:
                _run._terminate_run_if_running(termination_status)

        atexit.register(
            terminate_run_if_running_with_weakref, weakref.ref(self), RunStatus.FINISHED
        )
        if self._sender_job:
            self._sender_job.start()
        if self._system_metrics_job:
            self._system_metrics_job.start()

    def end(self):
        self._terminate_run_if_running(RunStatus.FINISHED)

    def _add_git_info(self, root_path: Optional[str] = None):
        root_path = root_path or os.getcwd()
        try:
            git_info = GitInfo(root_path)
            tags = [
                RunTag(
                    key=constants.GIT_COMMIT_TAG_NAME,
                    value=git_info.current_commit_sha,
                ),
                RunTag(
                    key=constants.GIT_BRANCH_TAG_NAME,
                    value=git_info.current_branch_name,
                ),
                RunTag(key=constants.GIT_DIRTY_TAG_NAME, value=str(git_info.is_dirty)),
            ]
            remote_url = git_info.remote_url
            if remote_url is not None:
                tags.append(RunTag(key=constants.GIT_REMOTE_URL_NAME, value=remote_url))
            self.mlflow_client.log_batch(run_id=self.run_id, tags=tags)
            log_artifact_blob(
                mlflow_client=self.mlflow_client,
                run_id=self.run_id,
                blob=git_info.diff_patch,
                file_name=constants.PATCH_FILE_NAME,
                artifact_path=constants.PATCH_FILE_ARTIFACT_DIR,
            )
        except Exception as ex:
            # no-blocking
            logger.warning(f"failed to log git info because {ex}")

    def download_artifact(self, path: str, dst_path: Optional[str] = None) -> str:
        """
        Download recursively an artifact file or directory from a run to a local directory if applicable, and
        return a local path for it.

        Args:
            path : Relative source path to the desired artifact.
            dst_path : Absolute path of the local filesystem destination directory to which to download the specified
                artifacts. This directory must already exist. If unspecified, the artifacts will either be downloaded to
                a new uniquely-named directory on the local filesystem or will be returned directly in the case of the
                LocalArtifactRepository.
        Returns:
            str: Local path of desired artifact
        """
        if dst_path is None:
            return self.mlflow_client.download_artifacts(self.run_id, path=path)
        elif os.path.isdir(dst_path):
            return self.mlflow_client.download_artifacts(
                self.run_id, path=path, dst_path=dst_path
            )
        else:
            raise MlFoundryException(
                f"Destination path {dst_path} should be an existing directory."
            )

    def log_artifact(self, local_path: str, artifact_path: Optional[str] = None):
        if artifact_path is not None:
            NAMESPACE.validate_namespace_not_used(path=artifact_path)
        if os.path.isfile(local_path):
            self.mlflow_client.log_artifact(
                self.run_id, local_path=local_path, artifact_path=artifact_path
            )
        elif os.path.isdir(local_path):
            self.mlflow_client.log_artifacts(
                self.run_id, local_dir=local_path, artifact_path=artifact_path
            )
        else:
            raise MlFoundryException(
                f"local path {local_path} should be an existing file or directory"
            )

    def log_dataset(
        self,
        dataset_name: str,
        features,
        predictions=None,
        actuals=None,
        only_stats: bool = False,
    ):
        """
        Log a dataset associated with a run. A dataset is a collection of features,
        predictions and actuals. Datasets are uniquely identified by the dataset_name
        under a run. They are immutable, once successfully logged, overwriting it is not allowed.

        Mixed types are not allowed in features, actuals and predictions. However, there can be
        missing data in the form of None, NaN, NA.

        :param dataset_name:    Name of the dataset. Dataset name should only contain letters,
                                numbers, underscores and hyphens.
        :type dataset_name: str
        :param features:        Features associated with this dataset.
                                This should be either pandas DataFrame or should be of a
                                data type which can be convered to a DataFrame.
        :param predictions:     Predictions associated with this dataset and run. This
                                should be either pandas Series or should be of a data type which
                                can be convered to a Series. This is an optional argument.
        :param actuals:         Actuals associated with this dataset and run. This
                                should be either pandas Series or should be of a data type which
                                can be convered to a Series. This is an optional argument.
        :param only_stats:      If True, then the dataset (features, predictions, actuals) is
                                not saved. Only statistics and the dataset schema will be
                                persisted. Default is False.
        :type only_stats: bool
        """
        self._dataset_module.log_dataset(
            dataset_name=dataset_name,
            features=features,
            predictions=predictions,
            actuals=actuals,
            only_stats=only_stats,
        )

    def get_dataset(self, dataset_name: str) -> Optional[DataSet]:
        """
        Returns the features, predictions, actuals associated with the dataset.
        If only_stats was set to True while logging the dataset,
        this will return None.

        :param dataset_name: Name of the dataset.
        :type dataset_name: str
        :rtype: typing.Optional[DataSet]
        """
        return self._dataset_module.get_dataset(dataset_name=dataset_name)

    def log_metrics(self, metric_dict: Dict[str, Any], step: int = 0):
        """Logs the metrics given metric_dict that has metric name as key and metric value as value
        Args:
            metric_dict (dict): metric_dict that has metric name as key and metric value as value
            step (int): step associated with the metrics present in metric_dict
        Examples:
        >> log_metric({'accuracy':0.91, 'f1_score':0.5})
        """
        # not sure about amplitude tracking here.
        # as the user can use this function in training loop
        # amplitude.track(amplitude.Event.LOG_METRICS)

        try:
            # mlfow_client doesn't have log_metrics api, so we have to use log_batch,
            # This is what internally used by mlflow.log_metrics
            timestamp = int(time.time() * 1000)
            metrics_arr = [
                Metric(key, value, timestamp, step=step)
                for key, value in metric_dict.items()
            ]
            if len(metrics_arr) == 0:
                raise MlflowException("Cannot log empty metrics dictionary")

            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=metrics_arr, params=[], tags=[]
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        logger.info("Metrics logged successfully")

    def log_params(self, param_dict: ParamsType):
        """Logs the parameter given param_dict that has param name as key and param value as value
        Args:
            param_dict (dict): param_dict that has param name as key and param value as value
        Examples:
        >> log_params({'learning_rate':0.01, 'n_epochs:10'})

        We can also pass argparse Namspace objects directly

        ```
        import argparse
        parser = argparse.ArgumentParser()
        parser.add_argument("-batch_size", type=int, required=True)
        args = parser.parse_args()
        log_params(args)
        ```
        """
        amplitude.track(amplitude.Event.LOG_PARAMS)

        try:
            # mlfowclient doesnt have log_params api, so we have to use log_batch,
            # This is what internally used by mlflow.log_params
            param_dict = process_params(param_dict)
            params_arr = [Param(key, str(value)) for key, value in param_dict.items()]

            if len(params_arr) == 0:
                raise MlflowException("Cannot log empty params dictionary")

            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=[], params=params_arr, tags=[]
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        logger.info("Parameters logged successfully")

    def set_tags(self, tags: Dict[str, Any]):
        """Logs the given tags. Converts tag values to strings.
        Args:
            tags (dict): string tag name as key and tag value as value
        Examples:
        >> set_tags({'nlp.framework': 'Spark NLP'})
        """
        amplitude.track(amplitude.Event.SET_TAGS)

        try:
            NAMESPACE.validate_namespace_not_used(names=tags.keys())
            tags_arr = [RunTag(key, str(value)) for key, value in tags.items()]
            self.mlflow_client.log_batch(
                run_id=self.run_id, metrics=[], params=[], tags=tags_arr
            )
        except MlflowException as e:
            raise MlFoundryException(e.message) from e
        logger.debug("Set tags successfully")

    def get_tags(self) -> Dict[str, str]:
        amplitude.track(amplitude.Event.GET_TAGS)

        run = self.mlflow_client.get_run(self.run_id)
        return run.data.tags

    def __compute_whylogs_stats(self, df):

        if not isinstance(df, pd.DataFrame):
            raise MlFoundryException(
                f"df is expexted to be pd.DataFrame but got {str(type(df))}"
            )

        profile_file_name = (
            "profile"
            + "_"
            + datetime.datetime.now().strftime(constants.TIME_FORMAT)
            + ".bin"
        )
        session = whylogs.get_or_create_session()
        profile = session.new_profile()
        profile.track_dataframe(df)
        profile.write_protobuf(profile_file_name)

        try:
            self.mlflow_client.set_tag(self.run_id, "whylogs", True)
            self.mlflow_client.log_artifact(
                self.run_id,
                profile_file_name,
                artifact_path=MlFoundryRun.S3_WHYLOGS_PATH,
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        if os.path.exists(profile_file_name):
            os.remove(profile_file_name)

    def auto_log_metrics(
        self,
        model_type: enums.ModelType,
        data_slice: enums.DataSlice,
        predictions: Collection[Any],
        actuals: Optional[Collection[Any]] = None,
        class_names: Optional[List[str]] = None,
        prediction_probabilities=None,
    ) -> ComputedMetrics:
        metrics_calculator = get_metrics_calculator_v2(model_type)
        metrics = metrics_calculator.compute_metrics(
            predictions=predictions,
            actuals=actuals,
            prediction_probabilities=prediction_probabilities,
            class_names=class_names,
        )
        metric_path = os.path.join(constants.ALM_ARTIFACT_PATH, data_slice.value)
        log_artifact_blob(
            mlflow_client=self.mlflow_client,
            run_id=self.run_id,
            blob=metrics.json(),
            file_name=constants.ALM_METRICS_FILE_NAME,
            artifact_path=metric_path,
        )
        return metrics

    # TODO (nikunjbjj): This function is too long. Need to be broken into smaller testable modules.
    def log_dataset_stats(
        self,
        df,
        data_slice: enums.DataSlice,
        data_schema: Schema,
        model_type: enums.ModelType,
        shap_values=None,
    ):
        data_slice = enums.DataSlice(data_slice)
        model_type = enums.ModelType(model_type)

        if not isinstance(df, pd.DataFrame):
            raise MlFoundryException(f"Expected pd.DataFrame but got {str(type(df))}")

        if not data_schema.actual_column_name:
            raise MlFoundryException(f"Schema.actual_column_name cannot be None")
        elif not data_schema.prediction_column_name:
            raise MlFoundryException(f"Schema.prediction_column_name cannot be None")
        elif data_schema.feature_column_names is None:
            raise MlFoundryException(f"Schema.feature_column_names cannot be None")
        elif not isinstance(data_schema.feature_column_names, list):
            raise MlFoundryException(
                f"data_schema.feature_column_names should be of type list, "
                f"cannot be {type(data_schema.feature_column_names)}"
            )

        self.__compute_whylogs_stats(df[set(data_schema.feature_column_names)])

        if model_type in [
            enums.ModelType.BINARY_CLASSIFICATION,
            enums.ModelType.MULTICLASS_CLASSIFICATION,
        ]:
            class_names = None
            prediction_col_dtype, actual_col_dtype = (
                df[data_schema.prediction_column_name].dtype,
                df[data_schema.actual_column_name].dtype,
            )
            if prediction_col_dtype == object and actual_col_dtype != object:
                raise MlflowException(
                    "Both predictions column and actual column has to be of same datatype, either string or number"
                )
            elif prediction_col_dtype != object and actual_col_dtype == object:
                raise MlflowException(
                    "Both predictions column and actual column has to be of same datatype, either string or number"
                )
            elif prediction_col_dtype == object and actual_col_dtype == object:
                actual_class_names = df[data_schema.actual_column_name].unique()
                prediction_class_name = df[data_schema.prediction_column_name].unique()
                class_names = sorted(
                    set(actual_class_names) | set(prediction_class_name)
                )
                df[data_schema.actual_column_name] = df[
                    data_schema.actual_column_name
                ].apply(lambda x: class_names.index(x))
                df[data_schema.prediction_column_name] = df[
                    data_schema.prediction_column_name
                ].apply(lambda x: class_names.index(x))

        unique_count_dict = {}
        if model_type in [
            enums.ModelType.BINARY_CLASSIFICATION,
            enums.ModelType.MULTICLASS_CLASSIFICATION,
        ]:
            unique_count_dict[data_schema.prediction_column_name] = np.unique(
                df[data_schema.prediction_column_name].to_list(), return_counts=True
            )
            unique_count_dict[data_schema.actual_column_name] = np.unique(
                df[data_schema.actual_column_name].to_list(), return_counts=True
            )
        elif model_type == enums.ModelType.REGRESSION:
            session = whylogs.get_or_create_session()
            profile = session.new_profile()
            profile.track_dataframe(
                df[[data_schema.actual_column_name, data_schema.prediction_column_name]]
            )
            unique_count_dict[
                constants.ACTUAL_PREDICTION_COUNTS
            ] = profile.flat_summary()["hist"]

        if data_schema.categorical_feature_column_names:
            for feature in data_schema.categorical_feature_column_names:
                unique_count_dict[feature] = np.unique(
                    df[feature].to_list(), return_counts=True
                )

        unique_count_name = "unique_count" + "_" + str(data_slice.value) + ".json"
        constants.RUN_STATS_FOLDER.mkdir(parents=True, exist_ok=True)
        unique_count_path = os.path.join(constants.RUN_STATS_FOLDER, unique_count_name)

        with open(unique_count_path, "w") as fp:
            json.dump(unique_count_dict, fp, cls=NumpyEncoder)

        schema_json_name = "schema" + "_" + str(data_slice.value) + ".json"
        schema_json_path = os.path.join(constants.RUN_STATS_FOLDER, schema_json_name)

        with open(schema_json_path, "w") as outfile:
            json.dump(data_schema.__dict__, outfile)

        # TODO (nikunjbjj): This class name could be referenced before assignment. We need to fix this.
        if (
            model_type
            in [
                enums.ModelType.BINARY_CLASSIFICATION,
                enums.ModelType.MULTICLASS_CLASSIFICATION,
            ]
            and class_names is not None
        ):
            class_names_path = f"class_names_{data_slice.value}.json"
            class_names_path = constants.RUN_STATS_FOLDER / class_names_path
            class_names_dict = {"class_names": class_names}
            with open(class_names_path, "w") as fp:
                json.dump(class_names_dict, fp)

        metrics_class = get_metrics_calculator_v1(model_type)

        if data_schema.prediction_probability_column_name:
            metrics_dict = metrics_class.compute_metrics(
                df[set(data_schema.feature_column_names)],
                df[data_schema.prediction_column_name].to_list(),
                df[data_schema.actual_column_name].to_list(),
                df[data_schema.prediction_probability_column_name].to_list(),
            )
        else:
            metrics_dict = metrics_class.compute_metrics(
                df[set(data_schema.feature_column_names)],
                df[data_schema.prediction_column_name].to_list(),
                df[data_schema.actual_column_name].to_list(),
            )
        # non-multi dimensional metrics
        metrics_dict_with_data_slice = {}

        for key in metrics_dict[constants.NON_MULTI_DIMENSIONAL_METRICS].keys():
            new_key = "pre_computed_" + key + "_" + str(data_slice.value)
            metrics_dict_with_data_slice[new_key] = metrics_dict[
                constants.NON_MULTI_DIMENSIONAL_METRICS
            ][key]

        if shap_values is not None:
            tag_key = "data_stats_and_shap_" + data_slice.value
        else:
            tag_key = "data_stats_" + data_slice.value

        self.mlflow_client.set_tag(self.run_id, "modelType", model_type.value)
        self.mlflow_client.set_tag(self.run_id, tag_key, True)
        self.log_metrics(metrics_dict_with_data_slice)

        constants.RUN_METRICS_FOLDER.mkdir(parents=True, exist_ok=True)
        multi_dimension_metric_file = (
            "pre_computed_"
            + constants.MULTI_DIMENSIONAL_METRICS
            + "_"
            + str(data_slice.value)
            + ".json"
        )
        multi_dimension_metric_file_path = os.path.join(
            constants.RUN_METRICS_FOLDER, multi_dimension_metric_file
        )

        with open(multi_dimension_metric_file_path, "w") as fp:
            json.dump(
                metrics_dict[constants.MULTI_DIMENSIONAL_METRICS], fp, cls=NumpyEncoder
            )

        if model_type == enums.ModelType.TIMESERIES:
            actuals_predictions_filename = (
                "actuals_predictions_" + str(data_slice.value) + ".parquet"
            )
            actuals_predictions_filepath = os.path.join(
                constants.RUN_STATS_FOLDER, actuals_predictions_filename
            )
            df[
                [data_schema.prediction_column_name, data_schema.actual_column_name]
            ].to_parquet(actuals_predictions_filepath)

        try:

            # with self.mlflow_run as run:
            self.mlflow_client.log_artifact(
                self.run_id, unique_count_path, artifact_path=MlFoundryRun.S3_STATS_PATH
            )
            self.mlflow_client.log_artifact(
                self.run_id,
                multi_dimension_metric_file_path,
                artifact_path=MlFoundryRun.S3_METRICS_PATH,
            )
            self.mlflow_client.log_artifact(
                self.run_id, schema_json_path, artifact_path=MlFoundryRun.S3_STATS_PATH
            )
            if (
                model_type
                in [
                    enums.ModelType.BINARY_CLASSIFICATION,
                    enums.ModelType.MULTICLASS_CLASSIFICATION,
                ]
                and class_names is not None
            ):
                self.mlflow_client.log_artifact(
                    self.run_id,
                    class_names_path,
                    artifact_path=MlFoundryRun.S3_STATS_PATH,
                )
            # TODO (nikunjbjj): This class name path and actuals_predictions_filepath could be referenced
            #  before assignment. We need to fix this.
            if model_type == enums.ModelType.TIMESERIES:
                self.mlflow_client.log_artifact(
                    self.run_id,
                    actuals_predictions_filepath,
                    artifact_path=MlFoundryRun.S3_STATS_PATH,
                )
                os.remove(actuals_predictions_filepath)
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        os.remove(multi_dimension_metric_file_path)
        os.remove(schema_json_path)
        os.remove(unique_count_path)

        if shap_values is not None:
            self.__log_shap_values(
                df[set(data_schema.feature_column_names)], shap_values, data_slice
            )

        logger.info("Dataset stats have been successfully computed and logged")

    def __log_shap_values(self, df, shap_values: list, data_slice: enums.DataSlice):

        if not isinstance(df, pd.DataFrame):
            raise MlFoundryException(f"Expected pd.DataFrame but got {str(type(df))}")

        artifact_name = str(self.run_id) + "_" + str(data_slice.value) + "_shap.json"
        constants.RUN_STATS_FOLDER.mkdir(parents=True, exist_ok=True)
        filename = os.path.join(constants.RUN_STATS_FOLDER, artifact_name)

        shap_values_dict = {"shap_values": shap_values}

        with open(filename, "w") as fp:
            json.dump(shap_values_dict, fp, cls=NumpyEncoder)

        try:
            self.mlflow_client.log_artifact(
                self.run_id, filename, artifact_path=MlFoundryRun.S3_STATS_PATH
            )
        except MlflowException as e:
            raise MlFoundryException(e.message).with_traceback(
                e.__traceback__
            ) from None

        os.remove(filename)

    def get_metrics(
        self, metric_names: Optional[Iterable[str]] = None
    ) -> Dict[str, List[Metric]]:
        """
        Returns logged metrics.

        Args:
            metric_names: a list of metric names. If not passed,
                          Then returns all metrics logged for the run.
        Returns:
            A dictionary of metric_names and the logger metrics.
            The key is metric name. Value is a list of mlflow.entities.Metric object.
            If a metric has not been used before, value will be an empty list.

        ```
        In [4]: test_run.log_metrics({"loss": 0.6}, step=1)
        2022-02-21 22:32:19.550 INFO    mlfoundry.mlfoundry_run: Metrics logged successfully

        In [5]: test_run.log_metrics({"loss": 0.3}, step=2)
        2022-02-21 22:32:27.191 INFO    mlfoundry.mlfoundry_run: Metrics logged successfully

        In [6]: test_run.get_metrics()
        Out[6]:
        {'loss': [<Metric: key='loss', step=1, timestamp=1645462939544, value=0.6>,
          <Metric: key='loss', step=2, timestamp=1645462947185, value=0.3>]}
        ```
        """

        amplitude.track(amplitude.Event.GET_METRICS)
        run = self.mlflow_client.get_run(self.run_id)
        run_metrics = run.data.metrics

        metric_names = (
            set(metric_names) if metric_names is not None else run_metrics.keys()
        )

        unknown_metrics = metric_names - run_metrics.keys()
        if len(unknown_metrics) > 0:
            logger.warning(f"{unknown_metrics} metrics not present in the run")
        metrics_dict = {metric_name: [] for metric_name in unknown_metrics}
        valid_metrics = metric_names - unknown_metrics
        for metric_name in valid_metrics:
            metrics_dict[metric_name] = self.mlflow_client.get_metric_history(
                self.run_id, metric_name
            )
        return metrics_dict

    def get_params(self):
        # Ending any run that is active to prevent mlflow from throwing error
        amplitude.track(amplitude.Event.GET_PARAMS)
        run = self.mlflow_client.get_run(self.run_id)
        return run.data.params

    def get_model(self, dest_path: Optional[str] = None, **kwargs):
        """
        Deserialize and return the logged model object.
        dest_path is used to download the model before it is deserialized.
        Apart from model_name and dest_path, all the other keyword args (kwargs)
        will be passed to the deserializer.

        :param dest_path: The path where the model is downloaded before deserializing.
        :type dest_path: typing.Optional[str]
        :param kwargs: Keyword arguments to be passed to the deserializer.
        """
        return self._model_driver.get_model(dest_path=dest_path, **kwargs)

    def download_model(self, dest_path: str):
        """
        Download logged model for a run in a local directory.
        If dest_path does not exist, a new directory will be created.
        If dest_path already exist, it should be an empty directory.

        :param dest_path: local directory where the model will be downloaded.
        :type dest_path: str
        """
        return self._model_driver.download_model(dest_path=dest_path)

    def log_model(self, model, framework: Union[enums.ModelFramework, str], **kwargs):
        """
        Serialize and log a model under a run. After logging, we
        cannot overwrite it.

        :param model: The model object
        :param framework: Model Framework. Ex:- pytorch, sklearn
        :type framework: str
        :param kwargs: Keyword arguments to be passed to the serializer.
        """
        self._model_driver.log_model(model=model, framework=framework, **kwargs)



import datetime
import os
from pathlib import Path

from mlfoundry.internal_namespace import NAMESPACE

DEFAULT_TRACKING_URI = (
    "https://mlf-server.tfy-ctl-euwe1-devtest.devtest.truefoundry.tech"
)

# TODO (nikhil): This entire servicefoundry path needs to be changed to MLF folder
SERVICE_FOUNDRY_FOLDER = Path(os.path.abspath("./servicefoundry"))
RUN_TMP_FOLDER = Path(
    os.path.join(SERVICE_FOUNDRY_FOLDER, "logdirs")
)  # this is the log directory

RUN_PREDICTIONS_FOLDER = Path(
    os.path.join(RUN_TMP_FOLDER, "predictions")
)  # path to store predictions
RUN_DATASET_FOLDER = Path(
    os.path.join(RUN_TMP_FOLDER, "datasets")
)  # path to store datasets
# path to store dataset stats
RUN_STATS_FOLDER = Path(os.path.join(RUN_TMP_FOLDER, "stats"))
RUN_METRICS_FOLDER = Path(os.path.join(RUN_TMP_FOLDER, "metrics"))

GET_RUN_TMP_FOLDER = Path(
    os.path.join(SERVICE_FOUNDRY_FOLDER, "getdirs")
)  # this is the log directory
GET_RUN_PREDICTIONS_FOLDER = Path(
    os.path.join(GET_RUN_TMP_FOLDER, "predictions")
)  # path to store predictions
GET_RUN_DATASET_FOLDER = Path(
    os.path.join(GET_RUN_TMP_FOLDER, "datasets")
)  # path to store datasets

TIME_LIMIT_THRESHOLD = datetime.timedelta(minutes=1)
FILE_SIZE_LIMIT_THRESHOLD = pow(10, 7)  # 10MB

MLFLOW_HOST = "http://localhost:5000"


MULTI_DIMENSIONAL_METRICS = "multi_dimensional_metrics"  # multi dimensional metrics
NON_MULTI_DIMENSIONAL_METRICS = (
    "non_multi_dimensional_metrics"  # non-multi dimensional metrics
)
PROB_MULTI_DIMENSIONAL_METRICS = (
    "prob_multi_dimensional_metrics"  # multi dimensional metrics
)
PROB_NON_MULTI_DIMENSIONAL_METRICS = (
    "prob_non_multi_dimensional_metrics"  # non-multi dimensional metrics
)


ACTUAL_PREDICTION_COUNTS = "actuals_predictions_counts"
MLF_FOLDER_NAME = "mlf"
MLRUNS_FOLDER = "mlruns"
MLRUNS_FOLDER_NAME = Path(os.path.join(MLF_FOLDER_NAME, MLRUNS_FOLDER))


# Runs Name and DF constants
RUN_ID_COL_NAME = "run_id"
RUN_NAME_COL_NAME = "run_name"

TIME_FORMAT = "%Y-%m-%d_%H-%M-%S"


GIT_COMMIT_TAG_NAME = NAMESPACE("git.commit_sha")
GIT_BRANCH_TAG_NAME = NAMESPACE("git.branch_name")
GIT_REMOTE_URL_NAME = NAMESPACE("git.remote_url")
GIT_DIRTY_TAG_NAME = NAMESPACE("git.dirty")
PATCH_FILE_ARTIFACT_DIR = NAMESPACE / "git"
LOG_DATASET_ARTIFACT_DIR = NAMESPACE / "datasets"
# without .txt patch file does not load on UI
PATCH_FILE_NAME = "uncommitted_changes.patch.txt"


ALM_SCALAR_METRICS_PATTERN = NAMESPACE("auto_logged_metrics.{data_slice}.{metric_name}")
ALM_ARTIFACT_PATH = NAMESPACE / "auto_logged_metrics"
ALM_METRICS_FILE_NAME = "metrics.json"



import atexit
import logging
import os
import queue
import threading
import typing

from mlfoundry.exceptions import MlFoundryException

from .constants import (
    FLUSH_BATCH_SIZE_ENV_VAR,
    FLUSH_INTERVAL_ENV_VAR,
    MAX_QUEUE_SIZE_ENV_VAR,
)
from .inference_packet import ActualPacket, InferencePacket
from .repositories import AbstractInferenceStore
from .worker import InferenceStoreWorker

if typing.TYPE_CHECKING:
    from .inference_packet import BasePacket

logger = logging.getLogger(__name__)


def _check_shutdown(func):
    def wrapper(self, *args, **kwargs):
        if self._is_shutdown:
            raise MlFoundryException("cannot execute after client is shutdown")
        return func(self, *args, **kwargs)

    return wrapper


class InferenceStoreClient:
    def __init__(
        self,
        inference_repository_maker: typing.Callable[..., AbstractInferenceStore],
    ):
        self._is_shutdown = False

        self.inference_store: AbstractInferenceStore = inference_repository_maker()

        max_queue_size = int(os.getenv(MAX_QUEUE_SIZE_ENV_VAR, "10000"))
        flush_interval = float(os.getenv(FLUSH_INTERVAL_ENV_VAR, "5"))
        flush_batch_size = int(os.getenv(FLUSH_BATCH_SIZE_ENV_VAR, "100"))

        if max_queue_size <= 0:
            raise MlFoundryException(
                f"{MAX_QUEUE_SIZE_ENV_VAR} should be a positive number"
            )

        self.task_queue: "queue.Queue[BasePacket]" = queue.Queue(max_queue_size)
        self.worker_terminate_event = threading.Event()

        if flush_interval <= 0:
            raise MlFoundryException(
                f"{FLUSH_INTERVAL_ENV_VAR} should be a positive number"
            )
        if flush_batch_size <= 0 or flush_batch_size > max_queue_size:
            raise MlFoundryException(
                f"{FLUSH_BATCH_SIZE_ENV_VAR} should be positive"
                f" and lower than {MAX_QUEUE_SIZE_ENV_VAR}"
            )

        self.worker: InferenceStoreWorker = InferenceStoreWorker(
            task_queue=self.task_queue,
            terminate_event=self.worker_terminate_event,
            inference_store=self.inference_store,
            flush_interval=flush_interval,
            flush_every_num_message=flush_batch_size,
        )
        atexit.register(self.shutdown)
        self.worker.start()

    def _put_in_queue(self, inference_packet: InferencePacket):
        try:
            self.task_queue.put_nowait(inference_packet)
        except queue.Full as ex:
            raise MlFoundryException(
                "task queue is full\n"
                f"current task queue length is {MAX_QUEUE_SIZE}\n"
                "consider increasing the task queue length using "
                f"{MAX_QUEUE_SIZE_ENV_VAR} environment variable"
            ) from ex

    @_check_shutdown
    def log_predictions(self, predictions: typing.List[InferencePacket]):
        for prediction in predictions:
            self._put_in_queue(prediction)

    @_check_shutdown
    def log_actuals(self, actuals: typing.List[ActualPacket]):
        for actual in actuals:
            self._put_in_queue(actual)

    def flush(self):
        logger.debug(
            f"flushing task queue, {self.task_queue.qsize()} items in the queue"
        )
        self.task_queue.join()
        logger.debug("task queue flushed")

    def shutdown(self):
        if self._is_shutdown:
            return
        logger.debug("shutting down worker and client")
        self._is_shutdown = True
        # NOTE: We initialize the inference store at first in the constructor
        # The task_queue, worker is defined later.
        # There is a chance that inference  store initialization will throw error,
        # in that case, shutdown will be called (__del__) but self.task_queue would not have
        # been initialized yet.
        if hasattr(self, "task_queue"):
            self.flush()
        if hasattr(self, "worker_terminate_event"):
            logger.debug("setting worker termination event")
            self.worker_terminate_event.set()
        if hasattr(self, "worker"):
            logger.debug("waiting for worker to terminate")
            self.worker.join()

    def __del__(self):
        self.shutdown()



import logging
import os
import typing

import pandas as pd

from mlfoundry.enums import FileFormat
from mlfoundry.exceptions import MlFoundryException

logger = logging.getLogger(__name__)


def save_dataframe(
    local_dir: str,
    entity_name: str,
    dataframe: pd.DataFrame,
) -> typing.Tuple[str, FileFormat]:
    file_format = FileFormat.PARQUET
    file_name = f"{entity_name}.{file_format.value}"
    file_path = os.path.join(local_dir, file_name)
    try:
        dataframe.to_parquet(file_path, index=False)
        return file_name, file_format
    except Exception as ex:
        logger.info(
            f"failed to log {entity_name} as {FileFormat.PARQUET} "
            f"due to {ex}, trying with {FileFormat.CSV}"
        )

    file_format = FileFormat.CSV
    file_name = f"{entity_name}.{file_format.value}"
    file_path = os.path.join(local_dir, file_name)
    dataframe.to_csv(file_path, index=False)

    return file_name, file_format


def load_dataframe(local_file_path: str) -> pd.DataFrame:
    _, file_ext = os.path.splitext(local_file_path)
    if len(file_ext) <= 1:
        raise MlFoundryException(
            f"cannot recognize file format, '{file_ext}' is not valid"
        )
    # ".csv" -> "csv"
    file_ext = file_ext[1:]
    file_format = FileFormat(file_ext)

    if file_format is FileFormat.CSV:
        return pd.read_csv(local_file_path)
    elif file_format is FileFormat.PARQUET:
        return pd.read_parquet(local_file_path)
    else:
        raise MlFoundryException(
            f"cannot load dataframe from {file_format} file format"
        )


def save_series(
    local_dir: str,
    entity_name: str,
    series: pd.Series,
) -> typing.Tuple[str, FileFormat]:
    dataframe = series.to_frame(name=series.name or entity_name)
    return save_dataframe(
        local_dir=local_dir,
        entity_name=entity_name,
        dataframe=dataframe,
    )


def load_series(local_file_path: str) -> pd.Series:
    dataframe = load_dataframe(local_file_path)
    return dataframe.iloc[:, 0]



from mlfoundry.enums import ModelType
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.metrics.v1.base_metrics import BaseMetrics
from mlfoundry.metrics.v1.binary_classification_metrics import (
    BinaryClassificationMetrics,
)
from mlfoundry.metrics.v1.multiclass_classification_metrics import (
    MultiClassClassificationMetrics,
)
from mlfoundry.metrics.v1.regression_metrics import RegressionMetrics
from mlfoundry.metrics.v1.timeseries_metrics import TimeSeriesMetrics

MODEL_TYPE_TO_METRIC_CLASS_MAP = {
    ModelType.BINARY_CLASSIFICATION: BinaryClassificationMetrics,
    ModelType.MULTICLASS_CLASSIFICATION: MultiClassClassificationMetrics,
    ModelType.REGRESSION: RegressionMetrics,
    ModelType.TIMESERIES: TimeSeriesMetrics,
}


def get_metrics_calculator(model_type: ModelType, *args, **kwargs) -> BaseMetrics:
    if model_type not in MODEL_TYPE_TO_METRIC_CLASS_MAP:
        raise MlFoundryException(f"metrics computer for {model_type} is not registerd")
    return MODEL_TYPE_TO_METRIC_CLASS_MAP[model_type](*args, **kwargs)



import logging
import typing
from datetime import datetime

import numpy as np
from pydantic import BaseModel, Field, constr, validator

from mlfoundry.exceptions import MlFoundryException

logger = logging.getLogger(__name__)

ValueType = typing.Dict[str, typing.Union[str, float, int, bool, np.ndarray]]
# TODO: The user may want to pass np.uint8, np.float32 values directly. Support these types too.


class BasePacket(BaseModel):
    class Config:
        arbitrary_types_allowed = True


class InferencePacket(BasePacket):
    model_name: constr(min_length=1, max_length=256)
    model_version: constr(min_length=1, max_length=64)
    inference_id: constr(min_length=16, max_length=36)

    features: ValueType
    predictions: ValueType

    raw_data: typing.Optional[ValueType] = None
    actuals: typing.Optional[ValueType] = None
    occurred_at: datetime = Field(default_factory=datetime.utcnow)

    @validator("features", "predictions")
    def check_dict_length(cls, value):
        if len(value) == 0:
            raise MlFoundryException("features, predictions must be an non-empty dict")
        return value


class ActualPacket(BasePacket):
    model_name: constr(min_length=1, max_length=256)
    inference_id: constr(min_length=16, max_length=36)

    actuals: ValueType

    @validator("actuals")
    def check_dict_length(cls, value):
        if len(value) == 0:
            raise MlFoundryException("actuals must be an non-empty dict")
        return value



import typing

from mlfoundry.log_types.pydantic_base import PydanticBase


class DatasetSchemaRunLog(PydanticBase):
    value: typing.Dict

    @staticmethod
    def get_log_type() -> str:
        return "dataset/schema"



import typing
from functools import lru_cache

from mlflow.entities import FileInfo
from mlflow.protos.service_pb2 import ListArtifacts, MlflowService
from mlflow.store.tracking import rest_store
from mlflow.utils.proto_json_utils import message_to_json
from mlflow.utils.rest_utils import (
    _REST_API_PATH_PREFIX,
    extract_api_info_for_service,
    http_request_safe,
)

from mlfoundry.tracking.entities import TenantInfo
from mlfoundry.tracking.entities.artifact_credentials import ArtifactCredential

_METHOD_TO_INFO = extract_api_info_for_service(MlflowService, _REST_API_PATH_PREFIX)


class TruefoundryRestStore(rest_store.RestStore):
    def list_artifacts(self, run_id, path) -> typing.List[FileInfo]:
        infos = []
        page_token = None
        while True:
            if page_token:
                json_body = message_to_json(
                    ListArtifacts(run_id=run_id, path=path, page_token=page_token)
                )
            else:
                json_body = message_to_json(ListArtifacts(run_id=run_id, path=path))
            response = self._call_endpoint(ListArtifacts, json_body)
            artifact_list = response.files
            # If `path` is a file, ListArtifacts returns a single list element with the
            # same name as `path`. The list_artifacts API expects us to return an empty list in this
            # case, so we do so here.
            if (
                len(artifact_list) == 1
                and artifact_list[0].path == path
                and not artifact_list[0].is_dir
            ):
                return []
            for output_file in artifact_list:
                artifact_size = None if output_file.is_dir else output_file.file_size
                infos.append(
                    FileInfo(output_file.path, output_file.is_dir, artifact_size)
                )
            if len(artifact_list) == 0 or not response.next_page_token:
                break
            page_token = response.next_page_token
        return infos

    def get_artifact_read_credentials(self, run_id, path) -> ArtifactCredential:
        host_cred = self.get_host_creds()
        response = http_request_safe(
            host_creds=host_cred,
            endpoint="/api/2.0/mlflow/artifacts/credentials-read",
            method="get",
            params={"run_id": run_id, "path": path},
        )

        response = response.json()

        artifact_credential = ArtifactCredential(
            run_id=response["run_id"],
            path=response["path"],
            signed_uri=response["signed_uri"],
        )
        return artifact_credential

    def get_artifact_write_credential(self, run_id, path) -> ArtifactCredential:
        host_cred = self.get_host_creds()
        response = http_request_safe(
            host_creds=host_cred,
            endpoint="/api/2.0/mlflow/artifacts/credentials-write",
            method="get",
            params={"run_id": run_id, "path": path},
        )

        response = response.json()

        artifact_credential = ArtifactCredential(
            run_id=response["run_id"],
            path=response["path"],
            signed_uri=response["signed_uri"],
        )
        return artifact_credential

    # NOTE: get_tenant_id needs to only execute once.
    # get_tenant_id has no arguments and the response will not
    # change overtime
    # As functools.cache is not present in python 3.7 and 3.8,
    # I am using functools.lru_cache.
    @lru_cache(maxsize=1)
    def get_tenant_info(self) -> TenantInfo:
        host_cred = self.get_host_creds()
        response = http_request_safe(
            host_creds=host_cred,
            endpoint="/api/2.0/mlflow/tenant-id",
            method="get",
        )
        response = response.json()
        return TenantInfo.parse_obj(response)



import logging
import os
import typing

from mlfoundry.tracking.auth_service import AuthService
from mlfoundry.tracking.truefoundry_rest_store import TruefoundryRestStore

TRACKING_TOKEN_ENV_VAR = "MLFLOW_TRACKING_TOKEN"
API_KEY_ENV_VAR = "MLF_API_KEY"

logger = logging.getLogger(__name__)


class Session:
    def __init__(self, auth_service: AuthService):
        self.auth_service: AuthService = auth_service

    def init_session(
        self,
        api_key: typing.Optional[str] = None,
    ):
        final_api_key = api_key or os.getenv(API_KEY_ENV_VAR)

        if final_api_key is None:
            # if API key is not present,
            # then take a look if MLFLOW_TRACKING_TOKEN itself has been set.
            # this will be used in sfy.
            existing_token = os.getenv(TRACKING_TOKEN_ENV_VAR, "")
            if existing_token:
                logger.info("API key is not present. Using existing tracking token")
                return
            logger.info(
                "Session was not set as api key was neither passed, not set via env var"
            )
            return

        token = self.auth_service.get_token(api_key=final_api_key)
        os.environ[TRACKING_TOKEN_ENV_VAR] = token



import logging

import numpy as np
import pandas as pd
from sklearn.metrics import (
    accuracy_score,
    cohen_kappa_score,
    confusion_matrix,
    dcg_score,
    f1_score,
    log_loss,
    ndcg_score,
    precision_recall_curve,
    roc_curve,
)

from mlfoundry.constants import (
    MULTI_DIMENSIONAL_METRICS,
    NON_MULTI_DIMENSIONAL_METRICS,
    PROB_MULTI_DIMENSIONAL_METRICS,
    PROB_NON_MULTI_DIMENSIONAL_METRICS,
)
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.metrics.v1.base_metrics import BaseMetrics

logger = logging.getLogger(__name__)


class BinaryClassificationMetrics(BaseMetrics):
    TYPE_OF_MODEL = "binary_classification"
    ACCURACY = "accuracy"
    F1_SCORE = "f1_score"
    COHEN_KAPPA_SCORE = "cohen_kappa_score"
    CONFUSION_MATRIX = "confusion_matrix"
    DCG_SCORE = "dcg_score"
    NDCG_SCORE = "ndcg_score"
    LOG_LOSS = "log_loss"
    ROC_CURVE = "roc_curve"
    PR_CURVE = "precision_recall_curve"

    def __init__(self):
        super().__init__()
        self.binary_classification_metrics = {
            NON_MULTI_DIMENSIONAL_METRICS: {
                self.ACCURACY: self.compute_accuracy,
                self.F1_SCORE: self.compute_f1_score,
                self.COHEN_KAPPA_SCORE: self.compute_cohen_kappa_score,
            },
            MULTI_DIMENSIONAL_METRICS: {
                self.CONFUSION_MATRIX: self.compute_confusion_matrix,
            },
            PROB_NON_MULTI_DIMENSIONAL_METRICS: {
                # self.DCG_SCORE: self.compute_dcg_score,
                # self.NDCG_SCORE: self.compute_ndcg_score,
                self.LOG_LOSS: self.compute_log_loss,
            },
            PROB_MULTI_DIMENSIONAL_METRICS: {
                self.ROC_CURVE: self.compute_roc_curve,
                self.PR_CURVE: self.compute_pr_curve,
            },
        }

    def compute_metrics(
        self, features_df, predictions, labels, pred_probabilities=None
    ):
        """
        Compute metrics for binary classification models.
        This method calls the parent class as well for any supported metrics.
        :param features_df:
        :param predictions:
        :param labels:
        :param pred_probabilities:
        :return:
            metrics_dict
        """

        metrics_from_parent = super().compute_metrics(
            features_df, predictions, labels, pred_probabilities
        )

        all_metrics = metrics_from_parent
        for metric in self.binary_classification_metrics[
            NON_MULTI_DIMENSIONAL_METRICS
        ].keys():
            all_metrics[NON_MULTI_DIMENSIONAL_METRICS][
                metric
            ] = self.binary_classification_metrics[NON_MULTI_DIMENSIONAL_METRICS][
                metric
            ](
                labels, predictions
            )
        for metric in self.binary_classification_metrics[
            MULTI_DIMENSIONAL_METRICS
        ].keys():
            all_metrics[MULTI_DIMENSIONAL_METRICS][
                metric
            ] = self.binary_classification_metrics[MULTI_DIMENSIONAL_METRICS][metric](
                labels, predictions
            )

        if pred_probabilities:
            for metric in self.binary_classification_metrics[
                PROB_NON_MULTI_DIMENSIONAL_METRICS
            ].keys():
                all_metrics[NON_MULTI_DIMENSIONAL_METRICS][
                    metric
                ] = self.binary_classification_metrics[
                    PROB_NON_MULTI_DIMENSIONAL_METRICS
                ][
                    metric
                ](
                    labels, pred_probabilities
                )
            for metric in self.binary_classification_metrics[
                PROB_MULTI_DIMENSIONAL_METRICS
            ].keys():
                all_metrics[MULTI_DIMENSIONAL_METRICS][
                    metric
                ] = self.binary_classification_metrics[PROB_MULTI_DIMENSIONAL_METRICS][
                    metric
                ](
                    labels, pred_probabilities
                )

        return all_metrics

    @staticmethod
    def compute_accuracy(y_true, y_pred):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_pred:
        :return:
        """
        return accuracy_score(y_true, y_pred)

    @staticmethod
    def compute_f1_score(y_true, y_pred):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_pred:
        :return:
        """
        return f1_score(y_true, y_pred)

    @staticmethod
    def compute_cohen_kappa_score(y_true, y_pred):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_pred:
        :return:
        """
        return cohen_kappa_score(y_true, y_pred)

    @staticmethod
    def compute_confusion_matrix(y_true, y_pred):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_pred:
        :return:
        """
        return confusion_matrix(y_true, y_pred)

    @staticmethod
    def compute_dcg_score(y_true, y_prob):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_prob:
        :return:
        """
        return dcg_score(y_true, y_prob)

    @staticmethod
    def compute_ndcg_score(y_true, y_prob):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_prob:
        :return:
        """
        return ndcg_score(y_true, y_prob)

    @staticmethod
    def compute_log_loss(y_true, y_prob):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_prob:
        :return:
        """
        return log_loss(y_true, y_prob)

    @staticmethod
    def compute_roc_curve(y_true, y_prob):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_prob:
        :return:
        """
        classes = np.unique(y_true)
        if len(classes) != 2:
            raise MlFoundryException(
                "given data doesn't belong to binary classification"
            )

        y_pred = np.asarray(y_prob)
        fpr, tpr, thresholds = roc_curve(y_true, y_pred[:, 0], pos_label=classes[0])
        return {
            "fpr": fpr,
            "tpr": tpr,
            "thresholds": thresholds,
        }

    @staticmethod
    def compute_pr_curve(y_true, y_prob):
        """
        Sklearn metric compute.
        :param y_true:
        :param y_prob:
        :return:
        """
        classes = np.unique(y_true)
        if len(classes) != 2:
            raise MlFoundryException(
                "given data doesn't belong to binary classification"
            )

        y_pred = np.asarray(y_prob)
        precision, recall, thresholds = precision_recall_curve(
            y_true, y_pred[:, 0], pos_label=classes[0]
        )
        return {
            "precision": precision,
            "recall": recall,
            "thresholds": thresholds,
        }



import click

from mlfoundry import get_client


@click.group(help="download model or artifacts")
def download():
    ...


@download.command(short_help="Download logged model")
@click.option(
    "--run_fqn",
    required=True,
    type=str,
    help="fqn of the run",
)
@click.option(
    "--path",
    type=click.Path(file_okay=False, dir_okay=True, exists=False),
    required=True,
    help="path where the model will be downloaded",
)
def model(run_fqn: str, path: str):
    """
    Download the logged model for a run.\n
    """
    client = get_client()
    run = client.get_run_by_fqn(run_fqn)
    run.download_model(dest_path=path)



import os
import posixpath
import typing

from mlflow.entities import FileInfo
from mlflow.store.artifact.artifact_repo import ArtifactRepository
from mlflow.utils.file_utils import (
    download_file_using_http_uri,
    relative_path_to_artifact_path,
)
from mlflow.utils.rest_utils import cloud_storage_http_request

from mlfoundry.tracking.entities.artifact_credentials import ArtifactCredential
from mlfoundry.tracking.truefoundry_rest_store import TruefoundryRestStore


class TruefoundryArtifactRepository(ArtifactRepository):
    def __init__(self, artifact_uri, rest_store: TruefoundryRestStore):
        self.artifact_uri = artifact_uri
        super().__init__(artifact_uri)

        self.rest_store: TruefoundryRestStore = rest_store

    @staticmethod
    def _extract_run_id(artifact_uri) -> str:
        # artifact_uri will be something like,
        # s3://<BUCKET>/<PATH>/<EXP_ID>/<RUN_ID>/artifacts
        run_id = artifact_uri.rstrip("/").split("/")[-2]
        return run_id

    def list_artifacts(self, path=None) -> typing.List[FileInfo]:
        run_id = self._extract_run_id(self.artifact_uri)
        artifacts = self.rest_store.list_artifacts(run_id=run_id, path=path)
        return artifacts

    def _signed_uri_upload_file(
        self, artifact_credential: ArtifactCredential, local_file: str
    ):
        if os.stat(local_file).st_size == 0:
            with cloud_storage_http_request(
                "put",
                artifact_credential.signed_uri,
                data="",
            ) as response:
                response.raise_for_status()
        else:
            with open(local_file, "rb") as file:
                with cloud_storage_http_request(
                    "put",
                    artifact_credential.signed_uri,
                    data=file,
                ) as response:
                    response.raise_for_status()

    def log_artifacts(self, local_dir, artifact_path=None):
        dest_path = artifact_path or ""
        for (root, _, file_names) in os.walk(local_dir):
            upload_path = dest_path
            if root != local_dir:
                rel_path = os.path.relpath(root, local_dir)
                rel_path = relative_path_to_artifact_path(rel_path)
                upload_path = posixpath.join(dest_path, rel_path)
            for file_name in file_names:
                local_file = os.path.join(root, file_name)
                self.log_artifact(local_file=local_file, artifact_path=upload_path)

    def log_artifact(self, local_file, artifact_path=None):
        dest_path = self.artifact_uri
        if artifact_path:
            dest_path = posixpath.join(dest_path, artifact_path)
        dest_path = posixpath.join(dest_path, os.path.basename(local_file))
        artifact_credential = self.rest_store.get_artifact_write_credential(
            run_id=self._extract_run_id(self.artifact_uri), path=dest_path
        )
        self._signed_uri_upload_file(artifact_credential, local_file)

    def _download_file(self, remote_file_path: str, local_path: str):
        abs_remote_file_path = posixpath.join(self.artifact_uri, remote_file_path)
        artifact_credential = self.rest_store.get_artifact_read_credentials(
            run_id=self._extract_run_id(self.artifact_uri), path=abs_remote_file_path
        )
        download_file_using_http_uri(
            http_uri=artifact_credential.signed_uri, download_path=local_path
        )
